(require 'dash)
(require 'json)

(defun uniquify-frame-names ()
  "Give all frames a unique numeric ID."
  (progn (mapc
          (lambda (it)
            ;; Preserve already saved names.
            (unless (frame-parameter it '--saved-name)
              ;; Save the frame name
              (set-frame-parameter it '--saved-name
                                   (frame-parameter it 'name)))
            (set-frame-parameter it 'name (frame-parameter it 'window-id)))
          (frame-list))))

(defun restore-frame-names ()
  "Restore frame names from their saved value."
  (mapc
   (lambda (it)
     (when (frame-parameter it '--saved-name)
       (set-frame-parameter it 'name
                            (let ((name (frame-parameter it '--saved-name)))
                              ;; We can't restore names of the form
                              ;; F<number>, see frame.c:3080: error
                              ;; ("Frame names of the form F<num> are
                              ;; usurped by Emacs");
                              (unless (string-match "^F[[:digit:]]*$" name)
                                name)))
       (set-frame-parameter it '--saved-name nil)))
   (frame-list)))

(defun collect-emacs-frames-from-sway-tree (tree)
  "Return frames belonging to this Emacs in parsed Sway tree TREE."
  (let ((next-tree (gethash "nodes" tree)))
    (if (and
         (zerop (length next-tree))
         (not (string= "workspace" (gethash "type" tree)))
         (eq
          (gethash "pid" tree)
          (emacs-pid)))
        tree ; Collect
      (-flatten ; Or recurse
       (mapcar
        (lambda (t2) (collect-emacs-frames-from-sway-tree t2))
        next-tree)))))

;; ! Eval this again and again 
(prog2
    ;; Rename all frames
    (progn (uniquify-frame-names)
	   (redisplay t))
    ;; Get list of containers from sway, and find our frames by pid.
    (mapcar
     (lambda (it) (gethash "name" it))
     (collect-emacs-frames-from-sway-tree
      (with-temp-buffer
        (call-process "swaymsg" nil (current-buffer) nil "-t" "get_tree")
        (goto-char (point-min))
        (json-parse-buffer :null-object nil :false-object nil))))
  ;; Rename back frames.
  (restore-frame-names))
